include get.e
--include w32tk.e
--with trace

global integer f_debug    -- debug file

global function ltrim(sequence s)
-- remove space chars from the beginning of a string
  integer l

  l = length(s)
  if l=0 then return "" end if
  for i=1 to l do
    if s[i] != 32 then
      return s[i..l]
    end if
  end for
  return ""
end function

global function rtrim(sequence s)
-- remove space chars from the end of a string
  integer l

  l = length(s)
  if l=0 then return "" end if
  for i=l to 1 by -1 do
    if s[i] != 32 then
      return s[1..i]
    end if
  end for
  return ""
end function

global function trim(sequence s)
-- remove space chars from both ends of a string
  return ltrim(rtrim(s))
end function

global function rightPad(sequence s, atom len)
--Return a string right-padded to length len

  if length(s) >= len then
    return s
  else
    return s & repeat(' ', len-length(s))
  end if
end function

global function leftPad(sequence s, atom len)
--Return a string left-padded to length len

  if length(s) >= len then
    return s
  else
    return repeat(' ', len-length(s)) & s
  end if
end function

global function right(sequence s, integer n)
--Return the last n characters of a string
  integer l

  l = length(s)
  if (l=0) or (n=0) then return "" end if
  if n >= l then return s end if
  return s[l-n+1..l]
end function

-- function string was developped by Irv Mullins
global function string(object x)
  atom is_string
  if sequence(x) then
    is_string = 1
    for i = 1 to length(x) do
      if sequence(x[i]) then
        is_string = 0
      else
-- original version of Irv Mullins
--        if (x[i] < ' ') and (x[i] != 9) then is_string = 0 end if
        if not find(x[i], {9, 10, 13}) and (x[i] < ' ') then is_string = 0 end if
      end if
    end for
  else
    is_string = 0
  end if
  return is_string
end function

global function printSpecialCars(sequence s)
-- convert tab and end of line characters
  sequence converted
  integer i
  
  converted = {}
  i = 1
  while i <= length(s) do
    if s[i] = 9 then
      converted = converted & "\\t"
    elsif (s[i] = 13) and (s[i+1] = 10) then  -- DOS end of line
      converted = converted & "\\n"
      i = i + 1
    elsif s[i] = 10 then  -- Unix end of line
      converted = converted & "\\n"
    else
      converted = converted & s[i]
    end if
    i = i + 1
  end while
  return converted
end function

global procedure AnalyzeSequence(object s, sequence variable, integer output)
-- detail the content of a sequence
  if atom(s) then return end if
  if string(s) then
    printf(output, "%s = '%s'\n", {variable, printSpecialCars(s)})
  else
    for i=1 to length(s) do
      if integer(s[i]) then
        printf(output, "%s[%d] = %d\n", {variable, i, s[i]})
      elsif atom(s[i]) then
        printf(output, "%s[%d] = %f\n", {variable, i, s[i]})
      else
        AnalyzeSequence(s[i], variable&sprintf("[%d]", i), output)
      end if
    end for
  end if
end procedure

global function strtok(sequence chaine, atom separateur)
-- split a sequence
  integer slen
  sequence resultat, s
  integer str

  resultat = {}
  slen = length(chaine)
  s = {}
  str = 0

  for i = 1 to slen do
    if (chaine[i] = '"') and (separateur != '"') then
      if str = 0 then str = 1 else str = 0 end if
    end if
    if str = 0 and chaine[i] = separateur then
--      if length(s) > 0 then
        resultat = append(resultat, trim(s))
        s = {}
--      end if
	else
	  s = s & chaine[i]
	end if
  end for
--  if length(s) > 0 then
    resultat = append(resultat, trim(s))
--    s = {}
--  end if

  return resultat
end function

global function tokenize(sequence tableau, atom separateur)
-- join sequences via separators
  sequence s

  if atom(tableau[1]) then
    s = {tableau[1]}
  else
    s = tableau[1]
  end if
  for i = 2 to length(tableau) do
    if atom(tableau[1]) then
      s &= {separateur}&{tableau[i]}
    else
      s &= {separateur}&tableau[i]
    end if
  end for
--  puts(f_debug, s&"\n")
  return s
end function

global function searchList(sequence list, integer initial_row, integer col, object item)
-- search an item in a column of an array from a specified row
-- return the number of the line containing the item
  for i = initial_row to length(list) do
    if atom(item) then
      if atom(list[i][col]) and (item = list[i][col]) then
        return i
      end if
    elsif string(item) then
      if string(list[i][col]) and (compare(item, printSpecialCars(list[i][col])) = 0) then
        return i
      end if
    end if
  end for
  return 0
end function

global function searchArray(sequence list, sequence item)
-- search following items in an array
  integer trouve
  
  if atom(item[1]) then
    item = {item}
  end if
  trouve = 0
  for i = 1 to length(list) do
    trouve = i
    for j = 1 to length(item) do
--      if length(item[j])>0 then
        if compare(item[j], list[i][j]) != 0 then
          trouve = 0
        end if
--      end if
    end for
    if trouve > 0 then
      exit
    end if
  end for
  return trouve
end function

global function str_to_atom(sequence chaine)
-- convert a string to an atom
  integer n
  sequence s

  if length(chaine) = 0 then
    return 0
  end if
  n = find(',', chaine)
  if n>0 then chaine[n] = '.' end if
  s = value(chaine)
  if s[1] = GET_SUCCESS then
    return s[2]
  else
    printf(f_debug, "Erreur de conversion: %s\n", {chaine})
    abort(1)
  end if
end function

global function atom_to_str(atom valeur, sequence format)
-- convert an atom to a formatted string
  integer n
  sequence s

  s = sprintf(format, valeur)
  n = find('.', s)
  if n>0 then s[n] = ',' end if
  return s
end function

global function fileExists(sequence fichier)
-- check if file exists
  integer f_in

  f_in = open(fichier, "r")
  if f_in = -1 then
    return 0
  else
    close(f_in)
    return 1
  end if
end function

global function ReadCSV(sequence fichier)
-- read a CSV file (separator ';')
-- return a sequence
  sequence section
  integer f_in
  object line

  section = {}
  f_in = open(fichier, "r")
  if f_in != -1 then
    line = gets(f_in)
    if atom(line) then
      return section
    end if
    while 1 do
      line = gets(f_in)
      if atom(line) then
        exit
      end if
      line = line[1..length(line)-1]
      if length(line)>0 then
        section = append(section, strtok(line, ';') )
      end if
    end while
    close(f_in)
  end if
--  AnalyseSequence(section, "section", f_debug)
  return section
end function

global procedure SaveCSV(sequence fichier, sequence section)
-- write a CSV file
  integer f_out, l
  sequence line

--  AnalyseSequence(section, "section", f_debug)
  f_out = open(fichier, "w")
  if f_out != -1 then
    for i = 1 to length(section) do
      l = length(section[i])
      line = {}
      if l > 0 then
        for j = 1 to l-1 do
          line = line&section[i][j]&";"
        end for
        line = line&section[i][l]&"\n"
      end if
      puts(f_out, line)
--  AnalyseSequence(line, "line", f_debug)
    end for
    close(f_out)
  end if
end procedure

global function round(atom a, integer dec)
  -- rounds atom a to n decimals
  -- ex: round(112.234, 1) = 112.2
  -- n may be negative
  -- ex: round(112, -2) = 100
  atom n
  
  n = power(10, dec)
  return (floor((n*a)+ 0.5) / n)
end function

global function insertInSequence(object o, sequence s, integer p)
  -- inserts object o in sequence s at position p
  -- returns modified sequence
  if (p<1) or (p>length(s)) then return s end if
  if p = 1 then
    return o & s
  else
    return s[1..p-1] & o & s[p..$]
  end if
end function

global function removeFromSequence(sequence s, integer p)
  -- deletes object at position p in sequence s
  if (p<1) or (p>length(s)) then return s end if
  if p = 1 then
    return s[2..$]
  elsif p = length(s) then
    return s[1..$-1]
  else
    return s[1..p-1] & s[p+1..$]
  end if
end function

global procedure hexDump(sequence s, integer output)
  puts(output, "\n\nAddr  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F\n")
  puts(output, "0000 ")
  for i = 1 to length(s) do
    puts(output, sprintf(" %02x", s[i]))
    if remainder(i, 16) = 0 then
      puts(output, sprintf("\n%04x ", floor(i/16)))
    end if
  end for
  puts(output, "\n\n")
end procedure
