-- database.e
-- v0.4
-- April 25, 2007
-- reduced number of global variables
-- reduced size of field# number: 8 bits instead of 16
-- reduced size of record# number: 32 bits instead of 64
-- New database signature: BDD0.3
-- added sorting with CChris's sort_by_nth_field
-- added insertRecord
-- all databases can be computed simultaneously

-- v0.3
-- April 6, 2007

include get.e
include file.e
include functions.e
include common.e
include by_field.e

constant
  FIELD_TYPE = 1,
  FIELD_NAME = 2,
  FIELD_SIZE = 3,
  FIELD_DEC  = 4,
  FIELD_IDX  = 5

global constant
  SIGN  = 0,    -- 6 bytes long signature
  UINT8 = 1,    -- unsigned byte (8 bits)
  SINT8 = 2,    -- signed byte (8 bits)
  UINT16 = 3,   -- unsigned word (16 bits)
  SINT16 = 4,   -- signed word (16 bits)
  UINT32 = 5,   -- unsigned integer (32 bits)
  SINT32 = 6,   -- signed integer (32 bits)
  UINT64 = 7,   -- unsigned long integer (64 bits)
  FlOAT32 = 8,  -- single floating point number (32 bits)
  FLOAT64 = 9,  -- double floating point number (64 bits)
  STRING = 10,  -- null terminated string (in a fixed size location)
  DATE = 11,    -- date/time
  MEMO = 12,    -- memo (string of non limited size, stored in a memo file)
  LINK = 13     -- string refering an external location
  
constant
  SIGNATURE = "BDD0.3"

global constant
  DB_NAME = 1,
  DB_ID = 2,
  MEM_ID = 3,
  DB_FIELDS = 4,
  NB_RECORDS = 5,
  RECORD_LENGTH = 6,
  DB_INDEX = 7,
  DB_CURRENT = 8

global sequence OpenDB
OpenDB = {}

global integer sortByField    -- indicates on which field do sorting
sortByField = 0     -- no sorting

function currentRecord(integer db)
  return OpenDB[db][DB_CURRENT]
end function

function firstRecord(integer db)
  return 1
end function

function lastRecord(integer db)
  return length(OpenDB[db][DB_INDEX])
end function

function nextRecord(integer db)
  if OpenDB[db][DB_CURRENT] >= length(OpenDB[db][DB_INDEX])-1 then
    return OpenDB[db][DB_CURRENT]
  else
    return OpenDB[db][DB_CURRENT]+1
  end if
end function

function previousRecord(integer db)
  if OpenDB[db][DB_CURRENT] <= 1 then
    return OpenDB[db][DB_CURRENT]
  else
    return OpenDB[db][DB_CURRENT]-1
  end if
end function

function currentRecordPosition(integer db)
  return OpenDB[db][DB_INDEX][currentRecord(db)][1]
end function

function firstRecordPosition(integer db)
  return OpenDB[db][DB_INDEX][firstRecord(db)][1]
end function

function lastRecordPosition(integer db)
  return OpenDB[db][DB_INDEX][lastRecord(db)][1]
end function

function nextRecordPosition(integer db)
  return OpenDB[db][DB_INDEX][nextRecord(db)][1]
end function

function previousRecordPosition(integer db)
  return OpenDB[db][DB_INDEX][previousRecord(db)][1]
end function

function getRecordPosition(integer db, integer n)
  return 21 + 30*length(OpenDB[db][DB_FIELDS]) + (n-1) * OpenDB[db][RECORD_LENGTH]
end function

-- go to the given record
-- position is always related to the index
global function goRecord(integer db, integer n)
  integer ok

  if OpenDB[db][NB_RECORDS] = 0 then
--    ok = seek(db, firstRecordPosition(db))
--    OpenDB[db][DB_CURRENT] = 0
    return 0
  end if
  OpenDB[db][DB_CURRENT] = n
  ok = seek(OpenDB[db][DB_ID], getRecordPosition(db, currentRecordPosition(db)))
  return 1
end function

-- go to the first record
-- position is always related to the index
global function goFirstRecord(integer db)
  return goRecord(db, 1)
end function

-- go to the next record
-- position is always related to the index
global function goNextRecord(integer db)
  return goRecord(db, nextRecord(db))
end function

-- go to the previous record
-- position is always related to the index
global function goPreviousRecord(integer db)
  return goRecord(db, previousRecord(db))
end function

-- go to the last record
-- position is always related to the index
global function goLastRecord(integer db)
  return goRecord(db, lastRecord(db))
end function

function getChecksum(integer db)
  integer ok
  sequence s

  ok = seek(db, 7)   -- go to ckecksum field
  -- get checksum
  s = getUInt64(db)
  if s[1] = GET_EOF then
    puts(1, "Corrupted file! No checksum field!\n")
    return 0
  end if
  return s[2]
end function

function setChecksum(integer db, atom cks)
  integer ok

  ok = seek(db, 7)   -- go to ckecksum field
  -- set checksum
  ok = putInt64(db, cks)
  return 1
end function

function calcChecksum(integer db)
  integer ok, c
  atom cks_calc

  ok = seek(OpenDB[db][DB_ID], 15)   -- go to encryption status field
  cks_calc = 0
  while 1 do
    c = getc(OpenDB[db][DB_ID])
    if c = -1 then
      exit
    else
      cks_calc += c
    end if
  end while
  return cks_calc
end function

-- recalculate checksum and update checksum field
-- go back to previous record
function updateChecksum(integer db)
  integer number, ok

  number = calcChecksum(db)
  ok = setChecksum(OpenDB[db][DB_ID], number)
  ok = seek(OpenDB[db][DB_ID], currentRecord(db))
  return 1
end function

-- calculate the checksum of the database file
global function verifyChecksum(integer db)
  atom cks_rec, cks_calc

  cks_rec = getChecksum(OpenDB[db][DB_ID])
  if cks_rec = 0 then
    puts(1, "Corrupted database!\n")
    return 0
  end if
  cks_calc = calcChecksum(db)
  if cks_calc != cks_rec then
    printf(1, "Wrong checksum!\nRecorded: %d\nCalculated: %d\n", {cks_rec, cks_calc})
    return 0
  end if
  return cks_calc
end function

function getLockingStatus(integer db)
  integer ok
  sequence s

  ok = seek(OpenDB[db][DB_ID], 6)   -- go to locking status field
-- get locking status
  s = getUInt8(OpenDB[db][DB_ID])
  if s[1] = GET_EOF then
    puts(1, "Corrupted file! No locking status field!\n")
    return 0
  end if
  return s[2]+1  -- because 0 means failed
end function

function setLockingStatus(integer db, integer val)
  integer ok

  ok = seek(OpenDB[db][DB_ID], 6)   -- go to locking status field
-- set locking status
  ok = putInt8(OpenDB[db][DB_ID], val)
  return 1
end function

global function createDatabase(sequence name, sequence fields, object key)
-- creates a database
-- if key is a sequence then the database is (will be) crypted
  integer ok, db, mem
  integer nb_fields, field_type
  atom checksum

  db = open(name&".bdd", "wb")  -- a new database file
  if db = -1 then
    puts(1, "Error opening database file!\n")
    return 0
  end if
  mem = open(name&".mem", "wb")  -- a new database file
  if mem = -1 then
    puts(1, "Error opening memo file!\n")
    return 0
  end if
-- header
  ok = putSignature(db, SIGNATURE)  -- signature, pos 0
  ok = putInt8(db, 0)  -- locking status, pos 6
  ok = putInt64(db, 0)   -- checksum, pos 7
  checksum = 0
  if atom(key) then
    checksum += putInt8(db, 0)  -- encryption status, pos 15
  else
    checksum += putInt8(db, 1)  -- encryption status, pos 15
  end if
  nb_fields = length(fields)
  checksum += putInt8(db, nb_fields)  -- number of fields, pos 16
  checksum += putInt32(db, 0)   -- number of records, pos 17
  for i = 1 to nb_fields do
    checksum += putInt8(db, fields[i][FIELD_TYPE])  -- field type
    checksum += putString(db, fields[i][FIELD_NAME], 25) -- field name
    field_type = fields[i][FIELD_TYPE]
    if field_type = SIGN then
      checksum += putInt16(db, 6)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = UINT8 then
      checksum += putInt16(db, 1)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = SINT8 then
      checksum += putInt16(db, 1)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = UINT16 then
      checksum += putInt16(db, 2)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = SINT16 then
      checksum += putInt16(db, 2)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = UINT32 then
      checksum += putInt16(db, 4)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = SINT32 then
      checksum += putInt16(db, 4)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = UINT64 then
      checksum += putInt16(db, 8)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = FlOAT32 then
      checksum += putInt16(db, 4)  -- field size
      checksum += putInt8(db, fields[i][FIELD_DEC])  -- field decimals number
    elsif field_type = FLOAT64 then
      checksum += putInt16(db, 8)  -- field size
      checksum += putInt8(db, fields[i][FIELD_DEC])  -- field decimals number
    elsif field_type = STRING then
      checksum += putInt16(db, fields[i][FIELD_SIZE])  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = DATE then
      checksum += putInt16(db, 7)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = MEMO then
      checksum += putInt16(db, 8)  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    elsif field_type = LINK then
      checksum += putInt16(db, fields[i][FIELD_SIZE])  -- field size
      checksum += putInt8(db, 0)   -- field decimals number
    else  -- unexpected type
      printf(1, "Unexpected type, correct database structure!\nType = %d\n", fields[i][FIELD_TYPE])
      close(db)
      close(mem)
      return 0
    end if
    checksum += putInt8(db, fields[i][FIELD_IDX])  -- field index status
  end for
-- records
  -- no record at this time
  -- records zone begins at position 21 + (nb_fields * 30)
  -- memo file is empty
-- checksum
  ok = setChecksum(db, checksum)
  close(db)
  close(mem)
  return 1
end function

function getNbFields(integer db)
  integer ok
  sequence s

  ok = seek(OpenDB[db][DB_ID], 16)   -- go to field number field
  s = getUInt8(OpenDB[db][DB_ID])
  if s[1] = GET_EOF then
    puts(1, "Corrupted file! No field number field\n")
    return 0
  end if
  return s[2]
end function

function getNbRecords(integer db)
  integer ok
  sequence s

  ok = seek(OpenDB[db][DB_ID], 17)   -- go to record number field
  s = getUInt32(OpenDB[db][DB_ID])
  if s[1] = GET_EOF then
    puts(1, "Corrupted file! No record number field\n")
    return 0
  end if
  return s[2]
end function

function setNbRecords(integer db, atom nb)
  integer ok

  ok = seek(OpenDB[db][DB_ID], 17)   -- go to record number field
  ok = putInt32(OpenDB[db][DB_ID], nb)
  return 1
end function

function getFields(integer db)
  integer ok, nb_fields
  sequence s
  sequence fields, field_desc

  nb_fields = getNbFields(db)
  ok = seek(OpenDB[db][DB_ID], 21)   -- go to field description zone
  fields = {}
-- get the fields description
  for i = 1 to nb_fields do
    field_desc = {0,"",0,0,0,0}
    -- get field type
    s = getUInt8(OpenDB[db][DB_ID])
    if s[1] = GET_EOF then
      printf(1, "Field #%d: No field type!\n", i)
      return 0
    end if
    field_desc[FIELD_TYPE] = s[2]
    -- get field name
    s = getString(OpenDB[db][DB_ID], 25)
    if s[1] = GET_EOF then
      printf(1, "Field #%d: No field name!\n", i)
      return 0
    end if
    field_desc[FIELD_NAME] = s[2]
    -- get field size
    s = getUInt16(OpenDB[db][DB_ID])
    if s[1] = GET_EOF then
      printf(1, "Field #%d: No field size!\n", i)
      return 0
    end if
    field_desc[FIELD_SIZE] = s[2]
    -- get field decimals number
    s = getSInt8(OpenDB[db][DB_ID])
    if s[1] = GET_EOF then
      printf(1, "Field #%d: No field decimals number!\n", i)
      return 0
    end if
    field_desc[FIELD_DEC] = s[2]
    -- get field index status
    s = getUInt8(OpenDB[db][DB_ID])
    if s[1] = GET_EOF then
      printf(1, "Field #%d: No field index status!\n", i)
      return 0
    end if
    field_desc[FIELD_IDX] = s[2]
    fields = append(fields, field_desc)
  end for
  return fields
end function

function getRecordLength(integer db)
  integer size
  sequence fields

  fields = getFields(db)
  size = 0
  for i = 1 to length(fields) do
    size += fields[i][FIELD_SIZE]
  end for
  return size
end function

-- read field #f of current record
global function getField(integer fn, sequence f)
  sequence s
  integer field_type

  field_type = f[FIELD_TYPE]
  if field_type = SIGN then
    s = getSignature(fn)
  elsif field_type = UINT8 then
    s = getUInt8(fn)
  elsif field_type = SINT8 then
    s = getSInt8(fn)
  elsif field_type = UINT16 then
    s = getUInt16(fn)
  elsif field_type = SINT16 then
    s = getSInt16(fn)
  elsif field_type = UINT32 then
    s = getUInt32(fn)
  elsif field_type = SINT32 then
    s = getSInt32(fn)
  elsif field_type = UINT64 then
    s = getUInt64(fn)
  elsif field_type = FlOAT32 then
    s = getFloat32(fn)
  elsif field_type = FLOAT64 then
    s = getFloat64(fn)
  elsif field_type = STRING then
    s = getString(fn, f[FIELD_SIZE])
  elsif field_type = DATE then
    s = getDate(fn)
  elsif field_type = MEMO then
    s = getUInt64(fn)
  elsif field_type = LINK then
    s = getString(fn, f[FIELD_SIZE])
  else  -- unexpected type
    printf(1, "Unexpected type, correct database structure!\nType = %d\n", f[FIELD_TYPE])
    return 0
  end if
  if s[1] = GET_EOF then
    printf(f_debug, "getField: No field %s!\n", {f[FIELD_NAME]})
    return 0
  end if
  return s[2]
end function

-- read current record
global function getRecord(integer db)
  sequence record, fields
  integer nb_fields, pos, ok

  pos = where(OpenDB[db][DB_ID])
  record = {}
  fields = OpenDB[db][DB_FIELDS]
  nb_fields = length(fields)
  for f = 1 to nb_fields do
    record = append(record, getField(OpenDB[db][DB_ID], OpenDB[db][DB_FIELDS][f]))
  end for
  ok = seek(OpenDB[db][DB_ID], pos)
  return record
end function

-- create index
-- return nb_records
global function createIndex(integer db)
  integer ok, n
  sequence idx, s

  OpenDB[db][DB_INDEX] = {}
  for i = 1 to OpenDB[db][NB_RECORDS] do
    n = getRecordPosition(db, i)
    ok = seek(OpenDB[db][DB_ID], getRecordPosition(db, i))
    s = getRecord(db)
    idx = {i}
    for f = 1 to length(OpenDB[db][DB_FIELDS]) do
      if OpenDB[db][DB_FIELDS][f][FIELD_IDX] = 1 then
        idx = append(idx, s[f])
      end if
    end for
    OpenDB[db][DB_INDEX] = append(OpenDB[db][DB_INDEX], idx)
  end for
--  AnalyzeSequence(OpenDB[db][DB_INDEX], sprintf("OpenDB[%db][DB_INDEX]", db), f_debug)
  OpenDB[db][DB_CURRENT] = 1
  return 1
end function

-- write 0s in the current record and remove it from the index
global function deleteRecord(integer db)
  integer ok, number
  sequence s

  s = repeat(0, OpenDB[db][RECORD_LENGTH])
  puts(OpenDB[db][DB_ID], s)
  if OpenDB[db][DB_CURRENT] = lastRecord(db) then OpenDB[db][DB_CURRENT] -= 1 end if
  OpenDB[db][DB_INDEX] = removeFromSequence(OpenDB[db][DB_INDEX], OpenDB[db][DB_CURRENT])
  number = OpenDB[db][NB_RECORDS]
  ok = setNbRecords(db, number-1)
  OpenDB[db][NB_RECORDS] = number-1
  ok = updateChecksum(db)
  ok = goRecord(db, OpenDB[db][DB_CURRENT])
  return 1
end function

-- write field #f of current record
global function writeField(integer fn, object o, sequence f, integer mem)
  integer ok, field_type
  atom number
  integer pos

  if fn = -1 then return 0 end if
  field_type = f[FIELD_TYPE]
  if field_type = SIGN then
    ok = putSignature(fn, o)
  elsif field_type = UINT8 then
    ok = putInt8(fn, o)
  elsif field_type = SINT8 then
    ok = putInt8(fn, o)
  elsif field_type = UINT16 then
    ok = putInt16(fn, o)
  elsif field_type = SINT16 then
    ok = putInt16(fn, o)
  elsif field_type = UINT32 then
    ok = putInt32(fn, o)
  elsif field_type = SINT32 then
    ok = putInt32(fn, o)
  elsif field_type = UINT64 then
    ok = putInt64(fn, o)
  elsif field_type = FlOAT32 then
    number = round(o, f[FIELD_DEC])
    ok = putFloat32(fn, number)
  elsif field_type = FLOAT64 then
    number = round(o, f[FIELD_DEC])
    ok = putFloat64(fn, number)
  elsif field_type = STRING then
    ok = putString(fn, o, f[FIELD_SIZE])
  elsif field_type = DATE then
    ok = putDate(fn, o[1][1], o[1][2], o[1][3], o[2][1], o[2][2], o[2][3])
  elsif field_type = MEMO then
    ok = seek(mem, -1)
    puts(mem, o)
    pos = where(mem)
    ok = putInt64(fn, pos)
  elsif field_type = LINK then
    ok = putString(fn, o, f[FIELD_SIZE])
  else  -- unexpected type
    printf(1, "Unexpected type, correct database structure!\nType = %d\n", f[FIELD_TYPE])
    return 0
  end if
  return 1
end function

-- sort the index by the specified field
-- uses CChris's sort_by_nth_field
-- sort may be increasing (direction>=0) or decreasing (direction<0)
global function sortOnField(integer db, integer field, integer direction)
  OpenDB[db][DB_INDEX] = sort_by_nth_field(OpenDB[db][DB_INDEX],field, direction)
  return 1
end function

-- write current record
global function writeRecord(integer db, sequence record)
  integer ok, nb_fields, l
  sequence fields

  if db = -1 then return 0 end if
  fields = OpenDB[db][DB_FIELDS]

-- add a byte at the beginning of a record
-- 1 meaning another record is following
-- 0 meaning this is the last record
-- this would allow repairing a database with a wrong checksum

-- think about backup copies of the database !

  nb_fields = length(fields)
  l = length(record)
  if l != nb_fields then
    printf(1, "Sequence length differs from field#!\nField# = %d\nRecord length = %d\n", {nb_fields, l})
    return 0
  end if
  for i = 1 to nb_fields do
    ok = writeField(OpenDB[db][DB_ID], record[i], fields[i], OpenDB[db][MEM_ID])
  end for
  return 1
end function

-- add a record to the database
global function addRecord(integer db, sequence record)
  integer ok, number
  sequence idx
  
  if db = -1 then return 0 end if

-- add a byte at the beginning of a record
-- 1 meaning another record is following
-- 0 meaning this is the last record
-- this would allow repairing a database with a wrong checksum

-- think about backup copies of the database !

  ok = seek(OpenDB[db][DB_ID], -1)   -- go to end of file
  ok = writeRecord(db, record)
  idx = {OpenDB[db][NB_RECORDS]+1}
  for f = 1 to length(OpenDB[db][DB_FIELDS]) do
    if OpenDB[db][DB_FIELDS][f][FIELD_IDX] = 1 then
      idx = append(idx, record[f])
    end if
  end for
  OpenDB[db][DB_INDEX] = append(OpenDB[db][DB_INDEX], idx)
  number = OpenDB[db][NB_RECORDS]
  ok = setNbRecords(db, number+1)
  OpenDB[db][NB_RECORDS] = number+1
  ok = updateChecksum(db)
  return 1
end function

-- modify current record
global function modifyRecord(integer db, sequence record)
  integer ok, idx
  sequence idx_rec

  idx = OpenDB[db][DB_CURRENT]
  ok = writeRecord(db, record)
  idx_rec = {OpenDB[db][DB_INDEX][currentRecord(db)][1]}
  for f = 1 to length(OpenDB[db][DB_FIELDS]) do
    if OpenDB[db][DB_FIELDS][f][FIELD_IDX] = 1 then
      idx_rec = append(idx_rec, record[f])
    end if
  end for
  OpenDB[db][DB_INDEX][currentRecord(db)] = idx_rec
  ok = updateChecksum(db)
  ok = goRecord(db, idx)
  return ok
end function

-- add the record at the end of the database and insert it in the index at the current position
global function insertRecord(integer db, sequence record)
  integer ok, idx, lg

  if OpenDB[db][DB_ID] = -1 then return 0 end if
  idx = OpenDB[db][DB_CURRENT]
  ok = addRecord(db, record)
  lg = length(OpenDB[db][DB_INDEX])
  OpenDB[db][DB_INDEX] = insertInSequence({OpenDB[db][DB_INDEX][lg]}, OpenDB[db][DB_INDEX], idx)
  OpenDB[db][DB_INDEX] = OpenDB[db][DB_INDEX][1..$-1]
  ok = updateChecksum(db)
  ok = goRecord(db, idx)
  return 1
end function

-- open the database
global function openDatabase(sequence name)
  integer ok, n, db, mem, currentDatabase
  atom cks_rec, cks_calc
  sequence fields

  n = searchList(OpenDB, 1, DB_NAME, name)
  if n > 0 then
    puts(1, "Database yet open!\n")
    return 0
  end if
  db = open(name&".bdd", "ub")  -- an existing database file
  if db = -1 then return 0 end if
  mem = open(name&".mem", "ub")  -- an existing memo file
  if mem = -1 then
    puts(1, "Corrupted database ! No memo file\n")
    close(db)
    return 0
  end if
  OpenDB = append(OpenDB, {name, db, mem, {}, 0, 0, {}, 0})
  currentDatabase = length(OpenDB)
  OpenDB[currentDatabase][NB_RECORDS] = getNbRecords(currentDatabase)
  fields = getFields(currentDatabase)
  if atom(fields) then
    puts(1, "Corrupted database ! No fields description\n")
    close(db)
    close(mem)
    return 0
  end if
  OpenDB[currentDatabase][DB_FIELDS] = fields
  OpenDB[currentDatabase][RECORD_LENGTH] = getRecordLength(currentDatabase)
  -- get checksum
  cks_rec = getChecksum(db)
  if cks_rec = 0 then
    puts(1, "Corrupted database! No checksum\n")
    close(db)
    close(mem)
    return 0
  end if
  cks_calc = calcChecksum(currentDatabase)
  if cks_calc != cks_rec then
    printf(1, "Wrong checksum!\nRecorded: %d\nCalculated: %d\n", {cks_rec, cks_calc})
    close(db)
    close(mem)
    return 0
  end if
  ok = setLockingStatus(currentDatabase, 1)     -- lock database
  OpenDB[currentDatabase][DB_CURRENT] = 0
  ok = createIndex(currentDatabase)
  ok = goFirstRecord(currentDatabase)   -- go to first record
  return currentDatabase
end function

-- close the database
global function closeDatabase(integer db)
  integer ok

  ok = setLockingStatus(db, 0)     -- unlock database
  close(OpenDB[db][DB_ID])
  close(OpenDB[db][MEM_ID])
  if length(OpenDB) = 1 then
    OpenDB = {}
    db = 0
    return 1
  end if
  if db = 1 then
    OpenDB = OpenDB[2..$]
  elsif db = length(OpenDB) then
    OpenDB = OpenDB[1..$-1]
    db = length(OpenDB)
  else
    OpenDB = OpenDB[1..db-1] & OpenDB[db+1..$]
  end if
  OpenDB[db][DB_CURRENT] = 0
  ok = goFirstRecord(db)   -- go to first record
  return db
end function

-- dumps a database in a file
-- if key is a sequence then the database will be decrypted (not available yet)
global function dumpDatabase(integer db, object key, integer output)
-- dumps a database in a file
-- if key is a sequence then the database will be decrypted
  integer ok, idx
  atom cks_rec
  integer encrypted, locked, nb_fields, nb_records
  sequence s
  sequence fields
  sequence records, result

  idx = OpenDB[db][DB_CURRENT]
  puts(output, "\nDump database\n\n")
  ok = seek(OpenDB[db][DB_ID], 0)         -- go to beginning of file
  result = {}
-- get signature
  s = getSignature(OpenDB[db][DB_ID])
  if s[1] = GET_EOF then
    puts(1, "Empty database!\n")
    return 0
  end if
  printf(output, "Signature: %s\n", {s[2]})
  if compare(s[2], SIGNATURE)!= 0 then
    puts(1, "Database not supported ! Wrong signature!\n")
    return 0
  end if

-- get locking status
  locked = getLockingStatus(db)
  if locked = 0 then
    puts(1, "Corrupted database! No locking status\n")
    return 0
  elsif locked = 1 then
    puts(output, "Database file not locked\n")
  else
    puts(output, "Database file locked\n")
  end if

  -- get checksum
  cks_rec = getChecksum(OpenDB[db][DB_ID])
  printf(output, "CheckSum: %d\n", cks_rec)
  if cks_rec = 0 then
    puts(1, "Corrupted database! Wrong checksum\n")
    return 0
  end if

-- get encryption status
  s = getUInt8(OpenDB[db][DB_ID])
  if s[1] = GET_EOF then
    puts(1, "Corrupted database! No encryption status!\n")
    return 0
  end if
  encrypted = s[2]
  if encrypted then
    puts(output, "Encrypted database file\n")
  else
    puts(output, "Database file not encrypted\n")
  end if

-- get the number of fields
  nb_fields = getNbFields(db)
  printf(output, "Number of fields: %d\n", nb_fields)
  if nb_fields = 0 then
    puts(1, "Corrupted database! No fields defined!\n")
    return 0
  end if

-- get the number of records
  nb_records = getNbRecords(db)
  printf(output, "Number of records: %d\n", nb_records)
  if nb_records = 0 then
    puts(1, "Corrupted database! No record number!\n")
    return 0
  end if

-- get the fields description
  fields = getFields(db)
  if atom(fields) then
    puts(1, "Corrupted database! No fields description\n")
    return 0
  end if
  for i = 1 to length(fields) do
    printf(output, "Field %d: name = %s, type = %d, size = %d, decimals = %d, indexed = %d\n",
      {i, fields[i][FIELD_NAME], fields[i][FIELD_TYPE], fields[i][FIELD_SIZE],
       fields[i][FIELD_DEC], fields[i][FIELD_IDX]})
  end for

  records = {}
-- get records
  for r = 1 to nb_records do
   ok = goRecord(db, r)
   records = append(records, getRecord(db))
  end for
  if nb_records = 0 then
    puts(1, "No record\n")
  else
    AnalyzeSequence(records, "records", output)
  end if

  ok = goRecord(db, idx)
  puts(output, "\nEnd of dump\n\n")
  return 1
end function

-- search the index for a specific item in the designed field
-- field is the number of the field in the index, not in the database
global function searchDatabase(integer db, integer field, object item)
  integer n, ok
  
  n = searchList(OpenDB[db][DB_INDEX], 1, field, item)
  if n > 0 then
    ok = goRecord(db, n)
    return getRecord(db)
  else
    return {}
  end if
end function

